local _, S = ...
local pairs, ipairs, string, type, time, GetTime = pairs, ipairs, string, type, time, GetTime

-- Default item sort
local function DefaultSort(entry1, entry2)
    if entry1.quality == entry2.quality then
        if entry1.effectiveILvl == entry2.effectiveILvl then
            if entry1.name == entry2.name then
                if entry1.count == entry2.count then
                    return entry1.bag * 36 + entry1.slot > entry2.bag * 36 + entry2.slot
                end
                return entry1.count > entry2.count
            end
            return entry1.name < entry2.name
        end
        return entry1.effectiveILvl > entry2.effectiveILvl
    end
    return entry1.quality > entry2.quality
end



-- ITEM GROUPS
-- All item lists should share this table for getting and setting the selected grouping method
S.itemGroupingSettings = {
    ["selectedGrouping"] = nil,
    ["collapsedGroups"] = {}
}
S.BINDING_GROUPS = {
    [0] = {
        ["name"] = S.Localize("SUBFILTER_BINDING_NONE"),
        ["order"] = 6,
    },
    [1] = {
        ["name"] = S.Localize("SUBFILTER_BINDING_ON_PICKUP"),
        ["order"] = 5,
    },
    [2] = {
        ["name"] = S.Localize("SUBFILTER_BINDING_ON_EQUIP").." ("..S.Localize("SUBFILTER_BINDING_SOULBOUND")..")",
        ["order"] = 2,
    },
    [3] = {
        ["name"] = S.Localize("SUBFILTER_BINDING_ON_USE").." ("..S.Localize("SUBFILTER_BINDING_SOULBOUND")..")",
        ["order"] = 4,
    },
    [4] = {
        ["name"] = S.Localize("SUBFILTER_BINDING_NONE"),
        ["order"] = 6,
    },
    [5] = {
        ["name"] = S.Localize("SUBFILTER_BINDING_ON_PICKUP"),
        ["order"] = 5,
    },
    [6] = {
        ["name"] = S.Localize("SUBFILTER_BINDING_ON_EQUIP"),
        ["order"] = 1,
    },
    [7] = {
        ["name"] = S.Localize("SUBFILTER_BINDING_ON_USE"),
        ["order"] = 3,
    }
}
-- Function 'func' takes an item and returns:
-- The name of the group
-- A number which is the order of the group when sorting
S.ItemGroups = {
    ["CATEGORY"] = {
        ["name"] = S.Localize("GROUPING_CATEGORY"),
        ["func"] = function(item)
            local categories = S.Settings.Get("categories2")
            for key, cat in pairs(categories) do
                if not S.Category.Filter(key, item) then
                    return cat.name, key
                end
            end
            return "None", 100
        end
    },
    ["FAVORITES"] = {
        ["name"] = S.Localize("FILTER_MARKER_ICON"),
        ["func"] = function(item)
            local markerIcon = S.Data.GetFavorited(item)
            if not markerIcon then
                markerIcon = 100
            end
            return S.Utils.FormatMarkerIcon(markerIcon), markerIcon
        end
    },
    ["TYPE"] = {
        ["name"] = S.Localize("COLUMN_TYPE"),
        ["func"] = function(item)
            if item.classID then
                return GetItemClassInfo(item.classID), item.classID
            else
                return "None", 100
            end
        end
    },
    ["QUALITY"] = {
        ["name"] = S.Localize("FILTER_QUALITY"),
        ["func"] = function(item)
            if item.quality then
                return S.Utils.GetItemQualityName(item.quality), -item.quality
            else
                return "None", 100
            end
        end
    },
    ["BINDING"] = {
        ["name"] = S.Localize("COLUMN_BINDING"),
        ["func"] = function(item)
            if not item.bindType then return S.BINDING_GROUPS[0].name, S.BINDING_GROUPS[0].order end
            local binding = item.bindType
            if not item.bound then binding = binding + 4 end
            return S.BINDING_GROUPS[binding].name, S.BINDING_GROUPS[binding].order
        end
    },
    ["EXPANSION"] = {
        ["name"] = S.Localize("COLUMN_EXPANSION"),
        ["func"] = function(item)
            if not item.expacID then return 0 end
            return S.Utils.FormatExpacLong(item.expacID), 18 - item.expacID
        end
    }
}



-- Item Columns
local function GetColumnWidth(self)
    return self.width
end
local function GetColumnMinWidthOrIconSize(self)
    local width = S.Settings.Get("iconSize") + 4
    if width > self.minWidth then
        return width
    else
        return self.minWidth
    end
end
S.ItemColumns = {
    ["FAVORITES"] = {
        ["name"] = S.Localize("COLUMN_FAVORITES"),
        ["width"] = 24,
        ["GetWidth"] = GetColumnWidth,
        ["align"] = "CENTER"
    },
    ["QUANTITY"] = {
        ["name"] = S.Localize("COLUMN_QUANTITY"),
        ["width"] = 42,
        ["GetWidth"] = GetColumnWidth,
        ["align"] = "RIGHT",
        ["sortMethods"] = {
            {
                ["title"] = "#",
                ["func"] = function(asc, slot1, slot2)
                    return S.Sort.ByKey(not asc, slot1, slot2, "count", DefaultSort)
                end
            }
        }
    },
    ["ICON"] = {
        ["name"] = S.Localize("COLUMN_ICON"),
        ["minWidth"] = 0,
        ["GetWidth"] = GetColumnMinWidthOrIconSize,
        ["align"] = "CENTER",
        ["sortMethods"] = {
            {
                ["title"] = "",
                ["func"] = function(asc, slot1, slot2) 
                    return S.Sort.ByKey(not asc, slot1, slot2, "quality", DefaultSort)
                end
            }
        }
    },
    ["NAME"] = {
        ["name"] = S.Localize("COLUMN_NAME"),
        ["GetWidth"] = function() return nil end,
        ["align"] = "CENTER",
        ["sortMethods"] = {
            {
                ["title"] = S.Localize("COLUMN_RARITY_SHORT"),
                ["func"] = function(asc, slot1, slot2) 
                    return S.Sort.ByKey(not asc, slot1, slot2, "quality", DefaultSort)
                end
            },
            {
                ["title"] = S.Localize("COLUMN_NAME_SHORT"),
                ["func"] = function(asc, slot1, slot2) 
                    return S.Sort.ByKey(asc, slot1, slot2, "name", DefaultSort)
                end,
                ["inverse"] = true
            }
        }
    },
    ["REQUIRED_LEVEL"] = {
        ["name"] = S.Localize("COLUMN_REQUIRED_LEVEL"),
        ["width"] = 38,
        ["GetWidth"] = GetColumnWidth,
        ["align"] = "CENTER",
        ["sortMethods"] = {
            {
                ["title"] = S.Localize("COLUMN_REQUIRED_LEVEL_SHORT"),
                ["func"] = function(asc, slot1, slot2)
                    -- Items without a level requirement can be either 1 or 0, however they are functionally the same
                    -- This sorting treats them as identical
                    local value1 = slot1.minLevel
                    local value2 = slot2.minLevel
                    if value1 < 1 then
                        value1 = 1
                    end
                    if value2 < 1 then
                        value2 = 1
                    end
                    return S.Sort.ByValue(not asc, value1, value2, slot1, slot2, DefaultSort)
                end
            }
        }
    },
    ["ITEM_LEVEL"] = {
        ["name"] = S.Localize("COLUMN_ITEM_LEVEL"),
        ["width"] = 38,
        ["GetWidth"] = GetColumnWidth,
        ["align"] = "CENTER",
        ["sortMethods"] = {
            {
                ["title"] = S.Localize("COLUMN_ITEM_LEVEL_SHORT"),
                ["func"] = function(asc, slot1, slot2) 
                    return S.Sort.ByKey(not asc, slot1, slot2, "effectiveILvl", DefaultSort)
                end
            }
        }
    },
    ["AGE"] = {
        ["name"] = S.Localize("COLUMN_TIME_ADDED"),
        ["width"] = 56,
        ["GetWidth"] = GetColumnWidth,
        ["align"] = "CENTER",
        ["sortMethods"] = {
            {
                ["title"] = S.Localize("COLUMN_TIME_ADDED_SHORT"),
                ["func"] = function(asc, slot1, slot2)
                    local age1, age2 = S.Data.GetItemAge(slot1), S.Data.GetItemAge(slot2)
                    return S.Sort.ByValue(asc, age1, age2, slot1, slot2, DefaultSort)
                end
            }
        }
    },
    ["TYPE"] = {
        ["name"] = S.Localize("COLUMN_TYPE_SHORT"),
        ["width"] = 80,
        ["GetWidth"] = GetColumnWidth,
        ["align"] = "CENTER",
        ["sortMethods"] = {
            {
                ["title"] = S.Localize("COLUMN_TYPE_SHORT"),
                ["func"] = function(asc, slot1, slot2)
                    return S.Sort.ByValue(asc, GetItemClassInfo(slot1.classID), GetItemClassInfo(slot2.classID), slot1, slot2, DefaultSort)
                end,
                ["inverse"] = true
            }
        }
    },
    ["SUBTYPE"] = {
        ["name"] = S.Localize("FILTER_SUBTYPE"),
        ["width"] = 80,
        ["GetWidth"] = GetColumnWidth,
        ["align"] = "CENTER",
        ["sortMethods"] = {
            {
                ["title"] = S.Localize("FILTER_SUBTYPE"),
                ["func"] = function(asc, slot1, slot2)
                    return S.Sort.ByValue(asc, GetItemSubClassInfo(slot1.classID, slot1.subClassID), GetItemSubClassInfo(slot2.classID, slot2.subClassID), slot1, slot2, DefaultSort)
                end,
                ["inverse"] = true
            }
        }
    },
    ["TYPE_ICON"] = {
        ["name"] = S.Localize("COLUMN_TYPE").." ("..S.Localize("COLUMN_ICON")..")",
        ["minWidth"] = 20,
        ["GetWidth"] = GetColumnMinWidthOrIconSize,
        ["align"] = "CENTER",
        ["sortMethods"] = {
            {
                ["title"] = "|TInterface\\Addons\\Sorted\\Textures\\Type-Icon-Heading:18:18:0:0:64:64:0:48:0:48|t",
                ["func"] = function(asc, slot1, slot2)
                    return S.Sort.ByValue(asc, GetItemClassInfo(slot1.classID), GetItemClassInfo(slot2.classID), slot1, slot2, DefaultSort)
                end
            }
        }
    },
    ["EXPANSION"] = {
        ["name"] = S.Localize("COLUMN_EXPANSION"),
        ["minWidth"] = 20,
        ["GetWidth"] = function(self)
            return GetColumnMinWidthOrIconSize(self) * 2
        end,
        ["align"] = "CENTER",
        ["sortMethods"] = {
            {
                ["title"] = S.Localize("COLUMN_EXPANSION_SHORT"),
                ["func"] = function(asc, slot1, slot2)
                    return S.Sort.ByKey(not asc, slot1, slot2, "expacID", DefaultSort)
                end
            }
        }
    },
    ["BINDING"] = {
        ["name"] = S.Localize("COLUMN_BINDING"),
        ["minWidth"] = 20,
        ["GetWidth"] = GetColumnMinWidthOrIconSize,
        ["align"] = "CENTER",
        ["sortMethods"] = {
            {
                ["title"] = "|TInterface\\Addons\\Sorted\\Textures\\Bind-Icons:0:0:0:0:128:64:38:56:7:28|t", --S.Localize("COLUMN_BINDING_SHORT"),
                ["func"] = function(asc, slot1, slot2)
                    -- Sorts by bindType, with unbound items first
                    local value1 = slot1.bindType * 2
                    local value2 = slot2.bindType * 2
                    if value1 == 0 then
                        value1 = 100
                    end
                    if value2 == 0 then
                        value2 = 100
                    end
                    if slot1.bound then 
                        value1 = value1 - 1
                    end
                    if slot2.bound then 
                        value2 = value2 - 1
                    end
                    return S.Sort.ByValue(not asc, value1, value2, slot1, slot2, DefaultSort)
                end
            }
        }
    },
    ["VALUE"] = {
        ["name"] = S.Localize("COLUMN_SELL_PRICE"),
        ["width"] = 48,
        ["GetWidth"] = GetColumnWidth,
        ["align"] = "CENTER",
        ["sortMethods"] = {
            {
                ["title"] = S.Localize("COLUMN_SELL_PRICE_SHORT"),
                ["func"] = function(asc, slot1, slot2) 
                    local value1 = slot1.value * slot1.count
                    local value2 = slot2.value * slot2.count
                    if slot1.hasNoValue then value1 = 0 end
                    if slot2.hasNoValue then value2 = 0 end
                    return S.Sort.ByValue(not asc, value1, value2, slot1, slot2, DefaultSort)
                end
            }
        }
    },
    ["TRASH"] = {
        ["name"] = S.Localize("COLUMN_TRASH"),
        ["width"] = 28,
        ["GetWidth"] = GetColumnWidth,
        ["align"] = "CENTER",
        ["sortMethods"] = {
            {
                ["title"] = "|TInterface\\Addons\\Sorted\\Textures\\Trash-Icons:0:0:0:0:128:64:38:58:5:25|t", --S.Localize("COLUMN_BINDING_SHORT"),
                ["func"] = function(asc, slot1, slot2)
                    local trash1, trash2 = S.Data.GetTrash(slot1), S.Data.GetTrash(slot2)
                    if slot1.hasNoValue then
                        trash1 = 3
                    elseif trash1 == 0 then
                        trash1 = 1
                    end
                    if slot2.hasNoValue then
                        trash2 = 3
                    elseif trash2 == 0 then
                        trash2 = 1
                    end
                    return S.Sort.ByValue(asc, trash1, trash2, slot1, slot2, DefaultSort)
                end
            }
        }
    }
}



local LIS = LibStub("LibItemSearch-1.2")
-- Item search is slow, only perform it occasionally while player is typing
local lastTimeSearched = GetTime()
local searchScheduled = false
local i = 1
local function DelayedFilter(self)
    lastTimeSearched = GetTime()
    local searchText = S.primaryFrame.searchBox:GetText()
    searchScheduled = false
    for _, entry in ipairs(self.entryData) do
        entry.filtered = false
        if #searchText > 0 then
            if not LIS:Matches(entry.data.link, searchText) then
                entry.filtered = true
            end
        end
        if not entry.filtered then
            entry.filtered = S.FilterItem(entry.data)
        end
    end
end
local function FilterEntries(self)
    local searchText = S.primaryFrame.searchBox:GetText()
    if #searchText > 0 and lastTimeSearched > GetTime() - 0.3 then
        if not searchScheduled then
            searchScheduled = true
            C_Timer.After(0.5, function() DelayedFilter(self) end)
        end
    else
        DelayedFilter(self)
        self:ScheduleUpdate(false, true)
    end
end


local function HasData(self)
    local data = S.Data.GetItem(self.bag, self.slot, S.GetSelectedCharacter())
    if data.link then return true end
    return false
end
local function EntryHasData(self, entry)
    return HasData(entry)
end
local function GetDataForEntry(self, entry)
    return S.Data.GetItem(entry.bag, entry.slot, S.GetSelectedCharacter())
end

local function EntryExists(self, entryDataIndex)
    if self.entryData[entryDataIndex] then
        local entryData = self.entryData[entryDataIndex]
        local itemData = S.Data.GetItem(entryData.bag, entryData.slot, S.GetSelectedCharacter())
        if itemData.link then
            return true
        end
    end
    return false
end

local function GetEntryFavorited(self, entryData)
    return S.Data.GetFavorited(entryData)
end


local freeSpaceColor, freeSpaceColorHighlight = CreateColor(0.96, 0.9, 0.82), CreateColor(1, 1, 1)
local freeSpaceLowColor, freeSpaceLowColorHighlight = CreateColor(1, 0.8, 0), CreateColor(1, 0.96, 0.5)
local freeSpaceZeroColor, freeSpaceZeroColorHighlight = CreateColor(1, 0.2, 0.3), CreateColor(1, 0.4, 0.6)
local freeSpaceMaxColor, freeSpaceMaxColorHighlight = CreateColor(0.96, 0.9, 0.82, 0.6), CreateColor(0.96, 0.9, 0.82, 0.9)
local function FreeSpaceTooltip(self)
    GameTooltip:SetOwner(self, "ANCHOR_LEFT")
    GameTooltip:ClearLines()

    local containers
    if self.type == "BAGS" then
        GameTooltip:AddLine(S.Localize("TOOLTIP_TITLE_BAGS"), 1, 1, 1, 1)
        containers = S.Utils.ContainersOfType("BAGS")
    elseif self.type == "BANK" then
        GameTooltip:AddLine(S.Localize("TOOLTIP_TITLE_BANK"), 1, 1, 1, 1)
        containers = S.Utils.ContainersOfType("BANK")
    elseif self.type == "REAGENT" then
        GameTooltip:AddLine(S.Localize("TOOLTIP_TITLE_REAGENTS"), 1, 1, 1, 1)
        containers = S.Utils.ContainersOfType("REAGENT")
    elseif self.type == "KEYRING" then
        GameTooltip:AddLine(KEYRING, 1, 1, 1, 1)
        containers = S.Utils.ContainersOfType("KEYRING")
    end
    GameTooltip:AddLine(" ")

    for i,container in pairs(containers) do
        local itemName, _, itemRarity

        if container ~= BANK_CONTAINER and container ~= BACKPACK_CONTAINER and container ~= REAGENTBANK_CONTAINER and container ~= KEYRING_CONTAINER then
            local invID = ContainerIDToInventoryID(container)
            local itemLink = GetInventoryItemLink("player", invID)
            if itemLink then
                itemName, _, itemRarity = GetItemInfo(itemLink)
            end
        else
            if self.type == "BAGS" then
                itemName, itemRarity = BACKPACK_TOOLTIP, 1
            elseif self.type == "BANK" then
                itemName, itemRarity = BANK, 1
            elseif self.type == "REAGENT" then
                itemName, itemRarity = REAGENT_BANK, 1
            elseif self.type == "KEYRING" then
                itemName, itemRarity = KEYRING, 1
            end
        end

        if itemName then
            local r,g,b = S.Utils.GetItemQualityColor(itemRarity)
            local r2, g2, b2 = 1, 1, 1
            if GetContainerNumFreeSlots(container) == 0 then
                r2, g2, b2 = 1, 0.2, 0.2
            elseif GetContainerNumSlots(container) * 0.1 >= GetContainerNumFreeSlots(container) then
                r2, g2, b2 = 1, 0.82, 0
            end
            GameTooltip:AddDoubleLine(itemName, (GetContainerNumSlots(container) - GetContainerNumFreeSlots(container)).."|cFFD1C9BF/"..GetContainerNumSlots(container), r, g, b, r2, g2, b2)
        end
    end

    if S.WoWVersion() >= 7 then
        GameTooltip:AddLine(" ")
        if self.type == "BAGS" then
            GameTooltip:AddLine(S.Localize("TOOLTIP_CLEANUP_BAGS"))
        elseif self.type == "BANK" then
            GameTooltip:AddLine(S.Localize("TOOLTIP_CLEANUP_BANK"))
        elseif self.type == "REAGENT" then
            GameTooltip:AddLine(S.Localize("TOOLTIP_CLEANUP_REAGENTS"))
        end
    end

    GameTooltip:Show()
end
local function FreeSpaceOnEnter(self)
    if self.numFreeSlots == 0 then
        self.text:SetTextColor(freeSpaceZeroColorHighlight:GetRGBA())
    elseif self.numFreeSlots <= self.numSlots * 0.05 then
        self.text:SetTextColor(freeSpaceLowColorHighlight:GetRGBA())
    else
        self.text:SetTextColor(freeSpaceColorHighlight:GetRGBA())
    end
    self.maxText:SetTextColor(freeSpaceMaxColorHighlight:GetRGBA())

    S.Tooltip.Schedule(function() FreeSpaceTooltip(self) end)
end
local function FreeSpaceOnLeave(self)
    if self.numFreeSlots == 0 then
        self.text:SetTextColor(freeSpaceZeroColor:GetRGBA())
    elseif self.numFreeSlots <= self.numSlots * 0.05 then
        self.text:SetTextColor(freeSpaceLowColor:GetRGBA())
    else
        self.text:SetTextColor(freeSpaceColor:GetRGBA())
    end
    self.maxText:SetTextColor(freeSpaceMaxColor:GetRGBA())

    S.Tooltip.Cancel()
end
local function FreeSpaceOnClick(self)
    if self.type == "BAGS" then
        SortBags()
    elseif self.type == "BANK" then
        SortBankBags()
    elseif self.type == "REAGENT" then
        SortReagentBankBags()
    end
end

local function SetMinimised(self, minimised)
    self.freeSpace:SetShown(not minimised)
    for k,v in pairs(self.containerButtons) do
        v:SetShown(not minimised)
    end
    if minimised then
        self:SetPoint("BOTTOMRIGHT")
    else
        self:SetPoint("BOTTOMRIGHT", 0, 32)
    end
end

local function IsAvailable(self)
    if not S.IsPlayingCharacterSelected() then
        return false
    end
    if self.type == "BAGS" or self.type == "KEYRING" then
        return true
    elseif self.type == "BANK" or self.type == "REAGENT" then
        return S.IsBankOpened()
    end
end

-- Replaces the OnUpdate of List. Adds filtering, and updating of the bag slot buttons
-- Only update when some time has passed since the last update
-- This variable is shared by all item lists, so they can't update on the same frame which distributes the lag
local lastUpdateTime = 0
local function OnUpdate3(self)
    if self.updateScheduled and GetTime() > lastUpdateTime + 0.05 then
        if self.resizeScheduled then
            self:OnResize()
        end
        self:UpdateColumnsSortArrows()
        self:UpdateColumns()
        if sort then
            self:UpdateDisplayedEntryData()
        end
        self:UpdateEntryButtons()
        self:UpdateScrollBarMax()
        self:UpdateContainerButtons()

        self.updateScheduled = false
        self.resizeScheduled = false
        self.sortingScheduled = false
        
        lastUpdateTime = GetTime()
    end
end
-- The OnUpdate function when the "Place item in here" overlay is showing
local function OnUpdate2(self)
    OnUpdate3(self)
    if not self:IsMouseOver() or not CursorHasItem() then
        self.placeItem:Hide()
        self:SetScript("OnUpdate", self.OnUpdate)
    end
end
local typeToStringTable = {
    ["BAGS"] = S.Localize("BAGS"),
    ["BANK"] = S.Localize("BANK"),
    ["REAGENT"] = S.Localize("REAGENTS"),
    ["KEYRING"] = KEYRING
}
local function TypeToString(type)
    return typeToStringTable[type]
end
-- The OnUpdate function when the "Place item in here" overlay isn't showing
-- Checks if the mouse is over and whether the overlay should be shown
local function OnUpdate(self)
    OnUpdate3(self)
    if self:IsAvailable() and self:IsMouseOver() then
        local infoType, itemID, itemLink = GetCursorInfo()
        if infoType == "item" then
            self.placeItem:Show()
            self.placeItem.text:SetText(string.format(S.Localize("BUTTON_PLACE_ITEM"), itemLink, TypeToString(self.type)))
            self:SetScript("OnUpdate", self.OnUpdate2)
        end
    end
end

local function UpdateEntryButtonContainerHighlights(self)
    for i = 1,self:GetNumVisibleEntries() do
        local v = self.entryButtons[i]
        if v:IsShown() then
            if v.bag == self.mouseIsOverContainer then
                v.containerHighlight:Show()
            else
                v.containerHighlight:Hide()
            end
        end
    end
end

local function AddContainerButton(self, containerID, gray)
    local parent = CreateFrame("FRAME", nil, self)
    local b = S.FrameTools.CreateCircleButton("Button", parent, not gray, nil, false)
    b.containerID = containerID
    b.inventoryID = ContainerIDToInventoryID(containerID)
    b:SetID(b.inventoryID)

    b:SetFrameLevel(parent:GetFrameLevel() + 1)
    b:SetPoint("LEFT", self.freeSpace, "RIGHT", 32 * #self.containerButtons, 0)

    b:RegisterForClicks("LeftButtonUp", "RightButtonUp")
    b:RegisterForDrag("LeftButton")
    b:SetScript("OnClick", function(self, button, down)
        if button == "LeftButton" then
            PickupInventoryItem(self.inventoryID)
        else
            S.Utils.EmptyBag(containerID)
        end
    end)
    b:SetScript("OnDragStart", function(self)
        PickupInventoryItem(self.inventoryID)
    end)
    b.UpdateTooltip = function(self)
        if S.IsPlayingCharacterSelected() and GetInventoryItemID("PLAYER", self.inventoryID) then
            GameTooltip:SetOwner(self, "ANCHOR_NONE")
            ContainerFrameItemButton_CalculateItemTooltipAnchors(self, GameTooltip)
            GameTooltip:SetInventoryItem("player", self.inventoryID)
            GameTooltip:AddLine(S.Localize("TOOLTIP_EMPTY_BAG"))
            GameTooltip:Show()
        end
    end
    
    b:SetScript("OnEnter", function(self)
        if self.list:IsAvailable() then
            self:UpdateTooltip()
        end
        self.list.mouseIsOverContainer = self.containerID
        self.list:UpdateEntryButtonContainerHighlights()
    end)
    b:SetScript("OnLeave", function(self)
        GameTooltip:Hide()
        self.list.mouseIsOverContainer = nil
        self.list:UpdateEntryButtonContainerHighlights()
    end)
    b:Show()
    b.list = self
    self.containerButtons[#self.containerButtons + 1] = b
end
local function UpdateContainerButtons(self)
    for k,v in pairs(self.containerButtons) do
        local itemData = S.Data.GetInventoryItem(v.inventoryID)
        if itemData then
            v.icon:SetTexture(itemData.texture)
        end
    end
end
local function UpdateContainerButtonsBank(self)
    local numSlots, full = GetNumBankSlots()
    if not S.IsPlayingCharacterSelected() then
        full = true
        numSlots = 10
    end
    if not full and S.IsBankOpened() then
        self.BuyBankSlotButton:Show()
        self.BuyBankSlotButton:SetPoint("LEFT", self.freeSpace, "RIGHT", numSlots * 32, 0)
        self.BuyBankSlotButton.text:SetText(S.Utils.FormatValueString(GetBankSlotCost(numSlots)))
    else
        self.BuyBankSlotButton:Hide()
    end
    for i,v in ipairs(self.containerButtons) do
        if i > numSlots then
            v:Hide()
        else
            local itemData = S.Data.GetInventoryItem(v.inventoryID)
            if itemData then
                v.icon:SetTexture(itemData.texture)
            end
            v:Show()
        end
    end
end

local function UpdateReagentWarningMessage(self)
    local data = S.GetData()
    if data.reagentNotUnlocked then
        if S.IsPlayingCharacterSelected() then
            if S.IsBankOpened() then
                self.reagentWarningMessage:Hide()
                self.reagentPurchaseButton:Show()
            else
                self.reagentWarningMessage:SetText(S.Localize("WARNING_REAGENTBANK_NOT_PURCHASED"))
                self.reagentWarningMessage:Show()
                self.reagentPurchaseButton:Hide()
            end
        else
            self.reagentWarningMessage:SetText(string.format(S.Localize("WARNING_REAGENTBANK_NOT_PURCHASED_OTHER"), data.name))
            self.reagentWarningMessage:Show()
            self.reagentPurchaseButton:Hide()
        end
    else
        self.reagentWarningMessage:Hide()
        self.reagentPurchaseButton:Hide()
    end
end

local function UpdateBankWarningMessage(self)
    local data = S.GetData()
    if data.bankNotCached then
        if S.IsPlayingCharacterSelected() then
            self.bankWarningMessage:SetText(S.Localize("WARNING_BANK_NOT_CACHED"))
        else
            self.bankWarningMessage:SetText(string.format(S.Localize("WARNING_BANK_NOT_CACHED_OTHER"), data.name))
        end
        self.bankWarningMessage:Show()
    else
        self.bankWarningMessage:Hide()
    end
end

local function UpdateEntryButtons(self)
    self:UpdateEntryButtonsSuper()
    self:UpdateEntryButtonContainerHighlights()
end

function S.CreateItemList(parent, type, itemButtonTemplate)
    local list =  S.CreateList(parent, S.CreateItemEntry, S.ItemColumns, "itemColumnSettings", true, S.ItemGroups, S.itemGroupingSettings)
    list:ClearAllPoints()
    list:SetPoint("TOPLEFT")
    list:SetPoint("BOTTOMRIGHT", 0, 32)

    list.canCombineStacks = true
    list.expandedCombinedItems = {}
    
    list.FilterEntries = FilterEntries
    list.EntryExists = EntryExists
    list.EntryHasData = EntryHasData
    list.GetDataForEntry = GetDataForEntry
    list.GetEntryFavorited = GetEntryFavorited
    list.itemButtonTemplate = itemButtonTemplate
    list.SetMinimised = SetMinimised
    list.IsAvailable = IsAvailable
    list.OnUpdate = OnUpdate
    list.OnUpdate2 = OnUpdate2
    list:SetScript("OnUpdate", OnUpdate)
    list.UpdateEntryButtonsSuper = list.UpdateEntryButtons
    list.UpdateEntryButtons = UpdateEntryButtons
    list.UpdateEntryButtonContainerHighlights = UpdateEntryButtonContainerHighlights

    list.AddContainerButton = AddContainerButton
    if type == "BANK" then
        list.UpdateContainerButtons = UpdateContainerButtonsBank
    else
        list.UpdateContainerButtons = UpdateContainerButtons
    end
    list.containerButtons = {}

    list.type = type
    list.containers = {}
    for k, bag in pairs(S.Utils.ContainersOfType(type)) do
        table.insert(list.containers, bag)
        for slot = 1, S.Utils.GetContainerMaxSlots(bag) do
            list:AddEntry({
                ["bag"] = bag, 
                ["slot"] = slot
            })
        end
    end

    if list.type == "BANK" then
        list.UpdateBankWarningMessage = UpdateBankWarningMessage
        list.bankWarningMessage = list:CreateFontString(nil, "OVERLAY", "SortedFont")
        list.bankWarningMessage:SetAllPoints()
        list.bankWarningMessage:SetTextScale(1.2)
        S.Utils.RunOnEvent(list, "CharacterSelected", list.UpdateBankWarningMessage)
        S.Utils.RunOnEvent(list, "BagsUpdated", list.UpdateBankWarningMessage)
    elseif list.type == "REAGENT" then
        list.UpdateReagentWarningMessage = UpdateReagentWarningMessage
        list.reagentWarningMessage = list:CreateFontString(nil, "OVERLAY", "SortedFont")
        list.reagentWarningMessage:SetAllPoints()
        list.reagentWarningMessage:SetTextScale(1.2)
        S.Utils.RunOnEvent(list, "CharacterSelected", list.UpdateReagentWarningMessage)
        S.Utils.RunOnEvent(list, "BagsUpdated", list.UpdateReagentWarningMessage)
        list.reagentPurchaseButton = CreateFrame("BUTTON", nil, list, "UIPanelButtonTemplate")
        local rpb = list.reagentPurchaseButton
        rpb:SetPoint("CENTER", -32, 0)
        rpb:SetSize(96, 32)
        rpb.text = rpb:CreateFontString(nil, "OVERLAY", "SortedFont")
        rpb:SetFontString(rpb.text)
        rpb:SetText(S.Localize("BUTTON_BUY_REAGENTS"))
        rpb.text1 = rpb:CreateFontString(nil, "OVERLAY", "SortedFont")
        rpb.text1:SetPoint("BOTTOM", rpb, "TOP", 0, 8)
        rpb.text1:SetTextScale(1.2)
        rpb.text1:SetPoint("LEFT", list)
        rpb.text1:SetPoint("RIGHT", list)
        rpb.text1:SetText(S.Localize("WARNING_BUY_REAGENTS"), 0, 10)
        rpb.text2 = rpb:CreateFontString(nil, "OVERLAY", "SortedFont")
        rpb.text2:SetPoint("LEFT", rpb, "RIGHT", 4, 0)
        rpb.text2:SetText(S.Utils.FormatValueString(GetReagentBankCost()))
        rpb.text2:SetTextScale(1.2)
        rpb:SetScript("OnClick", function(self) StaticPopup_Show("CONFIRM_BUY_REAGENTBANK_TAB") end)
        S.Utils.RunOnEvent(list, "ReagentsPurchased", list.UpdateReagentWarningMessage)

        list.depositReagentsButton = CreateFrame("BUTTON", nil, list)
        local drb = list.depositReagentsButton
        drb:SetNormalTexture("Interface\\Addons\\Sorted\\Textures\\Deposit-Reagents-Button")
        drb:SetHighlightTexture("Interface\\Addons\\Sorted\\Textures\\Deposit-Reagents-Button")
        drb:SetPushedTexture("Interface\\Addons\\Sorted\\Textures\\Deposit-Reagents-Button")
        drb:GetNormalTexture():SetTexCoord(0, 0.25, 0.25, 0.75)
        drb:GetHighlightTexture():SetTexCoord(0, 0.25, 0.25, 0.75)
        drb:GetPushedTexture():SetTexCoord(0.25, 0.5, 0.25, 0.75)
        drb:SetPoint("BOTTOM", parent, 0, -2)
        drb:SetSize(64, 32)
        drb:SetScript("OnEnter", function(self)
            S.Tooltip.CreateLocalized(self, "LEFT", "BUTTON_DEPOSIT_REAGENTS")
        end)
        drb:SetScript("OnLeave", function(self)
            S.Tooltip.Cancel()
        end)
        drb:SetScript("OnClick", function(self)
            PlaySound(SOUNDKIT.IG_MAINMENU_OPTION)
            DepositReagentBank()
        end)
        drb:SetScript("OnMouseDown", function(self)
            drb:GetHighlightTexture():SetTexCoord(0.25, 0.5, 0.25, 0.75)
        end)
        drb:SetScript("OnMouseUp", function(self)
            drb:GetHighlightTexture():SetTexCoord(0, 0.25, 0.25, 0.75)
        end)
        function drb:Update()
            self:SetShown(S.IsBankOpened())
        end
        S.Utils.RunOnEvent(drb, "BankOpened", drb.Update)
        S.Utils.RunOnEvent(drb, "BankClosed", drb.Update)
        drb:Hide()
    end

    S.Utils.RunOnEvent(list, "BagsUpdated", function(self)
        if self:IsShown() then
            self:ScheduleUpdate(false, true)
        end
    end)
    S.Utils.RunOnEvent(list, "LayoutChanged", function(self)
        self:UpdateEntryButtonContainerHighlights()
    end)
    S.Utils.RunOnEvent(list, "SearchChanged", function(self)
        if self:IsShown() then
            self:ScheduleUpdate(false, true)
            --self.scrollBar:SetValue(0)
        end
    end)
    S.Utils.RunOnEvent(list, "CategorySelected", function(self)
        if self:IsShown() then
            self:ScheduleUpdate(false, true)
            self:ScrollToTop()
        end
    end)
    S.Utils.RunOnEvent(list, "CharacterSelected", function(self)
        self.columnSettings = S.Settings.Get(self.columnSettingsKey)
        if self:IsShown() then
            self:ScheduleUpdate(false, true)
            self:ScrollToTop()
        end
    end)
    S.Utils.RunOnEvent(list, "EquipmentSetSelected", function(self)
        if self:IsShown() then
            self:ScheduleUpdate(false, true)
            self:ScrollToTop()
        end
    end)
    S.Utils.RunOnEvent(list, "FavoriteChanged", function(self)
        if self:IsShown() then
            self:ScheduleUpdate(false, true)
        end
    end)

    -- Update everything on show, since updates are disabled when the frame is hidden for performance.
    -- Also update BankFrame.selectedTab so items will go into either the bank or reagent bank correctly
    list:SetScript("OnShow", function(self)
        self:OnShow()
        if self.type == "BANK" then
            BankFrame.selectedTab = 1
            self:UpdateBankWarningMessage()
        elseif self.type == "REAGENT" then
            BankFrame.selectedTab = 2
            self:UpdateReagentWarningMessage()
        end
    end)


    list.placeItem = CreateFrame("BUTTON", nil, list)
    list.placeItem:SetPoint("TOPLEFT", list.head, "BOTTOMLEFT")
    list.placeItem:SetPoint("BOTTOMRIGHT")
    list.placeItem:SetFrameLevel(list:GetFrameLevel() + 100)
    list.placeItem:Hide()
    list.placeItem.bg = list.placeItem:CreateTexture(nil, "BACKGROUND")
    list.placeItem.bg:SetAllPoints()
    list.placeItem.bg:SetColorTexture(0, 0, 0, 0.5)
    S.FrameTools.AddBorder(list.placeItem, "highlight", "Interface\\Addons\\Sorted\\Textures\\Place-Item-Highlight", 8, 2, true)
    for k,v in pairs(list.placeItem.highlight.parts) do
        v:SetBlendMode("ADD")
    end
    list.placeItem.text = list.placeItem:CreateFontString(nil, "OVERLAY", "SortedFont")
    list.placeItem.text:SetTextColor(S.Utils.GetButtonTextColor():GetRGB())
    list.placeItem.text:SetPoint("TOPLEFT", 16, -16)
    list.placeItem.text:SetPoint("BOTTOMRIGHT", -16, 16)
    list.placeItem.text:SetTextScale(1.5)
    list.placeItem.list = list
    list.placeItem:SetScript("OnMouseDown", function(self)
        S.Utils.PlaceCursorItemInContainerType(self.list.type)
    end)

    
    list.freeSpace = CreateFrame("BUTTON", nil, list)
    list.freeSpace:SetPoint("TOPLEFT", list, "BOTTOMLEFT")
    list.freeSpace:SetSize(70, 32)
    list.freeSpace.text = list.freeSpace:CreateFontString(nil, "OVERLAY", "SortedFont")
    list.freeSpace.text:SetPoint("RIGHT", list.freeSpace, "CENTER", 0, 0)
    list.freeSpace.text:SetTextColor(freeSpaceColor:GetRGBA())
    list.freeSpace.text:SetTextScale(1.2)
    list.freeSpace.maxText = list.freeSpace:CreateFontString(nil, "OVERLAY", "SortedFont")
    list.freeSpace.maxText:SetPoint("BOTTOMLEFT", list.freeSpace.text, "BOTTOMRIGHT", 2, 0)
    list.freeSpace.maxText:SetTextColor(freeSpaceMaxColor:GetRGBA())
    list.freeSpace.parent = list
    list.freeSpace.type = type
    function list.freeSpace:Update()
        local data = S.GetData()
        if data.containerNumSlots then
            self.numSlots, self.numFreeSlots = 0, 0
            for k, bag in pairs(self.parent.containers) do
                if data.containerNumSlots[bag] then
                    self.numSlots = self.numSlots + data.containerNumSlots[bag].numSlots
                    self.numFreeSlots = self.numFreeSlots + data.containerNumSlots[bag].numFreeSlots
                end
            end
            if self.numSlots == 0 or self.type == "REAGENT" and data.reagentNotUnlocked then
                self.text:Hide()
                self.maxText:Hide()
            else
                self.text:Show()
                self.maxText:Show()
                self.text:SetText(self.numSlots - self.numFreeSlots)
                if self.numFreeSlots == 0 then
                    self.text:SetTextColor(freeSpaceZeroColor:GetRGBA())
                elseif self.numFreeSlots <= self.numSlots * 0.05 then
                    self.text:SetTextColor(freeSpaceLowColor:GetRGBA())
                else
                    self.text:SetTextColor(freeSpaceColor:GetRGBA())
                end
                self.maxText:SetText(self.numSlots)
            end
        else
            self.text:Hide()
            self.maxText:Hide()
        end
    end
    S.Utils.RunOnEvent(list.freeSpace, "BagsUpdated", list.freeSpace.Update)
    S.Utils.RunOnEvent(list.freeSpace, "CharacterSelected", list.freeSpace.Update)
    list.freeSpace:SetScript("OnEnter", FreeSpaceOnEnter)
    list.freeSpace:SetScript("OnLeave", FreeSpaceOnLeave)
    list.freeSpace:SetScript("OnClick", FreeSpaceOnClick)


    if type == "BANK" then
        list.BuyBankSlotButton = CreateFrame("BUTTON", nil, list)
        local b = list.BuyBankSlotButton
        b:SetSize(28, 28)
        b:SetNormalTexture("Interface\\Addons\\Sorted\\Textures\\CommonButtonsDropdown")
        b:GetNormalTexture():SetTexCoord(0, 0.375, 0, 0.375)
        b:SetHighlightTexture("Interface\\Addons\\Sorted\\Textures\\Close-Button-Highlight")
        b:SetPushedTexture("Interface\\Addons\\Sorted\\Textures\\CommonButtonsDropdown")
        b:GetPushedTexture():SetTexCoord(0.375, 0.75, 0, 0.375)
        b.text = b:CreateFontString(nil, "OVERLAY", "SortedFont")
        b.text:SetPoint("LEFT", b, "RIGHT")
        b.text:SetTextColor(1, 0.92, 0.8)
        b:SetScript("OnClick", PurchaseSlot)
        list:RegisterEvent("PLAYERBANKBAGSLOTS_CHANGED")
        list:RegisterEvent("PLAYERBANKSLOTS_CHANGED")
        list:HookScript("OnEvent", function(self, event)
            if event == "PLAYERBANKBAGSLOTS_CHANGED" or event == "PLAYERBANKSLOTS_CHANGED" then
                self:UpdateContainerButtons()
            end
        end)
        S.Utils.RunOnEvent(list, "BankOpened", list.UpdateContainerButtons)
        S.Utils.RunOnEvent(list, "BankClosed", list.UpdateContainerButtons)
        b:SetScript("OnEnter", function(self)
            S.Tooltip.CreateLocalized(self, "LEFT", "TOOLTIP_BUY_BANK_SLOT")
        end)
        b:SetScript("OnLeave", S.Tooltip.Cancel)
    end



    return list
end